from datetime import datetime
import logging
import os
import re

import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd

from . import _depchase, _fetchrepodata, _repodata
from .config import config
from .util import parse_dataset_name, rpm_name_only


log = logging.getLogger(__name__)


def _unwrap_description(desc):
    def replace(m):
        if len(m.group(0)) == 1:
            return ' '
        else:
            return m.group(0)[1:]

    return re.sub(r'\n+', replace, desc)


def _categorise_deps(all_rpm_deps):
    module_deps = set()
    remaining_rpm_deps = set()
    for pkgname in all_rpm_deps:
        modname = _repodata.get_module_for_rpm(pkgname)
        if modname is not None:
            module_deps.add(modname)
        else:
            remaining_rpm_deps.add(pkgname)
    return module_deps, remaining_rpm_deps


class ModuleGenerator(object):
    # If True, packages from the command line will be put into api: in the generated modulemd
    include_api_section = True
    # Rationale for packages from the command line
    api_component_rationale = "Package in api"
    # Rationale for other packages
    other_component_rationale = "Runtime dependency"

    def __init__(self, pkgs):
        self.pkgs = pkgs
        self.mmd = Modulemd.Module(mdversion=2)
        self._pool = _depchase.make_pool()
        self.rpm_branch = None

    def _calculate_dependencies(self):
        pkgs = self.pkgs
        pool = self._pool
        self.api_srpms = {
            rpm_name_only(_depchase.get_srpm_for_rpm(pool, dep))
            for dep in pkgs
        }
        run_deps = _depchase.ensure_installable(pool, pkgs)
        module_run_deps, rpm_run_deps = _categorise_deps(run_deps)
        # The platform module provides any other runtime dependencies - we
        # expect we'll always depend on it, but check for the heck of it.
        if len(rpm_run_deps) > 0:
            module_run_deps.add('platform')
        # We default to 'all available streams' for all dependencies we've
        # found. The module author may then edit this to exclude streams or
        # restrict to specific streams.
        self.module_run_deps = {m: [] for m in module_run_deps}
        self.run_srpms = set()

    def _save_module_md(self, output_fname):
        """
        Function writes modulemd to either stdout or the given filename
        :return:
        """
        if output_fname is not None:
            self.mmd.dump(output_fname)
            print('Generated modulemd file: %r' % output_fname)
        else:
            print(self.mmd.dumps())
        return True

    def _update_module_md(self):
        """
        Function updates modulemd file with dependencies
        are information taken from SPEC file.
        :return:
        """
        self.mmd.peek_module_licenses().add("MIT")

        # If we're generating a module, use the summary and description from
        # the first package. This isn't fully correct, but gives the module
        # maintainer a starting place to edit from.
        metadata = _depchase.get_rpm_metadata(self._pool, self.pkgs[0])
        self.mmd.props.summary = metadata['summary'].strip()
        self.mmd.props.description = _unwrap_description(
            metadata['description'].strip())

        # Declare the public API
        for pkg in self.api_srpms:
            if self.include_api_section:
                self.mmd.peek_rpm_api().add(pkg)

            component = Modulemd.ComponentRpm(
                name=pkg, rationale=self.api_component_rationale,
                buildorder=self._get_build_order(pkg))
            if self.rpm_branch is not None:
                component.props.ref = self.rpm_branch
            self.mmd.add_rpm_component(component)

        # Declare module level dependencies
        dependencies = Modulemd.Dependencies()
        for modname, streams in self.module_run_deps.items():
            dependencies.add_buildrequires(modname, streams)
            dependencies.add_requires(modname, streams)
        self.mmd.add_dependencies(dependencies)

        for pkg in self.run_srpms:
            component = Modulemd.ComponentRpm(
                name=pkg, rationale=self.other_component_rationale,
                buildorder=self._get_build_order(pkg))
            if self.rpm_branch is not None:
                component.props.ref = self.rpm_branch
            self.mmd.add_rpm_component(component)

        # TODO: Always set content licenses appropriately

        # TODO: Emit something for non-empty self.unresolved_build_rpms
        #       rather than relying solely on the warnings emitted on stderr

    def _get_build_order(self, pkg):
        if pkg in self.api_srpms:
            return 10
        else:
            return 0

    def run(self, output_fname):
        self._calculate_dependencies()
        self._update_module_md()
        self._save_module_md(output_fname)


class ModuleFromRepoGenerator(ModuleGenerator):

    def __init__(self, repodata_path, apipkgs, in_file=None):
        self.repodata_path = repodata_path
        self.apipkgs = apipkgs
        self.in_file = in_file

        self.install_local_repo()

        self.artifact_pkgs = self.get_artifact_pkgs()

        super().__init__(pkgs=())

        if in_file:
            self.mmd = Modulemd.Module.new_from_file(in_file)

    def __del__(self):
        # remove cache files we generated before
        for cache in _fetchrepodata._ALL_CACHES:
            cpath = _fetchrepodata._cache_fname(self.distropaths, cache)
            try:
                os.remove(cpath)
            except Exception:
                pass

    @property
    def distropaths(self):
        if not hasattr(self, '_distropaths'):
            relname, arch = parse_dataset_name()
            self._distropaths = _fetchrepodata.DistroPaths('_localrepo', arch)

            # fix local cache paths, we haven't downloaded this to the cache hierarchy
            for repopathspair in self._distropaths.repo_paths_by_name.values():
                for t in ('arch', 'src'):
                    getattr(repopathspair, t).local_metadata_path = self.repodata_path

        return self._distropaths

    def install_local_repo(self):
        # peek arch from dataset
        relname, arch = parse_dataset_name()

        # remove canned release configuration and replace with one containing
        # only the local repository
        config['datasets']['releases'] = {
            '_localrepo': {
                'default-architecture': arch,
                'repositories': {
                    '_everything': {
                        'arch': {'baseurl': f'file://{self.repodata_path}'},
                        'source': {'baseurl': f'file://{self.repodata_path}'},
                    },
                },
                '_localrepodatapath': self.repodata_path,
            },
        }

        _repodata.set_dataset_name('_localrepo')

        _fetchrepodata._write_lookup_caches(self.distropaths, silent=True)

    def get_artifact_pkgs(self):
        pkgs = []

        for repopathspair in self.distropaths.repo_paths_by_name.values():
            pkgs.extend(_fetchrepodata._read_packages(repopathspair.arch))

        return pkgs

    def _calculate_dependencies(self):
        # skip this step
        pass

    def _update_module_md(self):
        if not self.in_file:
            self.mmd.peek_module_licenses().add("MIT")

            # We're flying blind here, working only with the repo which is a tiny
            # subset of potentially available packages (or modules). Just add the
            # platform module as a dependency.
            dependencies = Modulemd.Dependencies()
            dependencies.add_buildrequires('platform', [])
            dependencies.add_requires('platform', [])
            self.mmd.add_dependencies(dependencies)

        if not self.in_file:
            rpm_api = self.mmd.peek_rpm_api()
            default_profile = Modulemd.Profile(name='default')
        else:
            rpm_api = None
            default_profile = None

        name_sum_desc_pkg = None

        for apipkg in self.apipkgs:
            for pkg_dct in self.artifact_pkgs:
                if pkg_dct['name'] == apipkg:
                    if not self.in_file and not name_sum_desc_pkg:
                        # set summary and description from API pkg
                        name_sum_desc_pkg = pkg_dct

                    if rpm_api:
                        rpm_api.add(apipkg)

                    if default_profile:
                        # only include profile block if API is specified and we're not
                        # modifying an existing modulemd file
                        default_profile.add_rpm(apipkg)

                    break
            else:
                log.warn(f"Didn't find artifact for API package {apipkg!r}.")

        if self.apipkgs and default_profile:
            self.mmd.add_profile(default_profile)

        if not self.in_file and not name_sum_desc_pkg:
            # default to first discovered package (if no API pkgs)
            try:
                name_sum_desc_pkg = self.artifact_pkgs[0]
                log.warn("Taking name/summary/description from first artifact package"
                         f" {name_sum_desc_pkg['name']!r}.")
            except IndexError:
                name_sum_desc_pkg = {
                    'name': "please-specify-a-module-name",
                    'summary': "A module",
                    'description': "A module generated from a plain RPM repository.",
                }
                log.warn("No artifact packages! Using boilerplate summary/description.")

        if name_sum_desc_pkg:
            self.mmd.props.name = name_sum_desc_pkg['name']
            self.mmd.props.summary = name_sum_desc_pkg['summary']
            self.mmd.props.description = _unwrap_description(
                name_sum_desc_pkg['description'].strip()
            )

        if not self.in_file:
            # set dummy stream name & context
            self.mmd.props.stream = 'please-set-a-stream'
            self.mmd.props.context = '00000000'

        # set module version to current time
        self.mmd.props.version = int(datetime.now().strftime('%Y%m%d%H%M%S'))


        if self.artifact_pkgs:
            # set artifacts
            artifacts_ss = Modulemd.SimpleSet()
            artifacts_ss.set([x['nevra'] for x in self.artifact_pkgs])
            self.mmd.set_rpm_artifacts(artifacts_ss)
