# -*- coding: utf-8 -*-
#
# sources.core - core functionality for accessing metadata from various sources
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from fnmatch import fnmatchcase
from typing import Iterator, Optional, Sequence, Tuple

import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd


class Source:
    """A source of information about modules, related artifacts, etc.

    This source can be a view on multiple "backends"."""

    _types_to_cls = {}

    def __init_subclass__(cls):
        stype = getattr(cls, 'source_type', None)
        if stype and stype not in Source._types_to_cls:
            Source._types_to_cls[stype] = cls

    @classmethod
    def from_type(cls, stype: str, *args, **kwargs):
        return cls._types_to_cls[stype](*args, **kwargs)

    @staticmethod
    def split_nsvc_fields(nsvc: str) -> Tuple[Optional[str]]:
        """Splits a module NSVC string into the corresponding 4-tuple.

        Missing or empty fields become None.

        :param nsvc: the NSVC string to split

        :returns:    a tuple of strings or None"""

        nsvc_fields = tuple(x if x else None for x in nsvc.split(":"))

        if len(nsvc_fields) > 4:
            raise ValueError(f"NSVC can only have 4 components: {nsvc!r}")

        nsvc_fields += (None,) * (4 - len(nsvc_fields))

        return nsvc_fields

    @staticmethod
    def filter_nsvc_fields(nsvc_fields_pats_list, nsvc_fields):
        filtered = []

        for nsvc_fields_pats in nsvc_fields_pats_list:
            for fp, f in zip(nsvc_fields_pats, nsvc_fields):
                if fp and f and not fnmatchcase(f, fp):
                    break
            else:
                filtered.append(nsvc_fields_pats)

        return filtered

    def enumerate_module_names(
        self,
        module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[str]:
        raise NotImplementedError()

    def enumerate_module_builds(
        self,
        nsvc_patterns: Optional[Sequence[str]] = None,
        latest: bool = True
    ) -> Iterator[Modulemd.Module]:
        raise NotImplementedError()

    def enumerate_module_defaults(
        self, module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[Modulemd.Defaults]:
        raise NotImplementedError()
