# -*- coding: utf-8 -*-
#
# sources.rpmrepo - functionality for accessing metadata from RPM repositories
#
# Copyright © 2018, 2019 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from fnmatch import fnmatchcase
import logging
from typing import Iterator, Optional, Sequence

import gi
gi.require_version('Modulemd', '1.0')  # noqa: E402
from gi.repository import Modulemd
import yaml

from .core import Source
from .._repodata import get_merged_modulemds
from .._repodata import get_modules_default_profiles_lookup
from .._repodata import get_modules_default_streams_lookup
from .._repodata import list_modules
from ..util import DefaultDictWithKey


log = logging.getLogger(__name__)


class RPMRepoDataSource(Source):

    source_type = 'rpmrepodata'

    @staticmethod
    def _name_matches_module_patterns(name, module_patterns):
        if not module_patterns:
            return True

        for pattern in module_patterns:
            if fnmatchcase(name, pattern):
                return True

        return False

    def enumerate_module_names(
        self,
        module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[str]:
        log.debug(f"enumerate_module_names({module_patterns!r})")

        return (
            x
            for x in list_modules()
            if self._name_matches_module_patterns(x, module_patterns)
        )

    def enumerate_module_builds(
        self,
        nsvc_patterns: Optional[Sequence[str]] = None,
        latest: bool = True
    ) -> Iterator[Modulemd.Module]:
        log.debug(f"enumerate_module_builds({nsvc_patterns!r}, {latest!r})")

        if nsvc_patterns:
            nsvc_patterns_fields = [self.split_nsvc_fields(pat)
                                    for pat in nsvc_patterns]
        else:
            nsvc_patterns_fields = None

        log.debug(f"\tnsvc_patterns_fields: {nsvc_patterns_fields!r}")

        for mmd in get_merged_modulemds():
            module = Modulemd.Module.new_from_string(yaml.dump(mmd))
            p = module.props
            nsvc = (p.name, p.stream, str(p.version), p.context)
            log.debug(f"nsvc: {nsvc!r}")
            if not nsvc_patterns_fields or self.filter_nsvc_fields(nsvc_patterns_fields, nsvc):
                yield module

    def enumerate_module_defaults(
        self, module_patterns: Optional[Sequence[str]] = None
    ) -> Iterator[Modulemd.Defaults]:
        log.debug(f"enumerate_module_defaults({module_patterns!r}")

        defaults = DefaultDictWithKey(
            lambda name: Modulemd.Defaults(module_name=name))

        for mname, dstream in get_modules_default_streams_lookup().items():
            if dstream:
                defaults[mname].set_default_stream(dstream)

        for mname, streams_dprofiles in get_modules_default_profiles_lookup().items():
            for stream, dprofiles in streams_dprofiles.items():
                defaults[mname].set_profiles_for_stream(stream, dprofiles)

        for mname, defaults in sorted(defaults.items(), key=lambda n_d: n_d[0]):
            if not module_patterns or any(fnmatchcase(mname, pat)
                                          for pat in module_patterns):
                yield defaults
